from coopr.pyomo import *
from coopr.opt import SolverFactory
from coopr.pyomo.base.expr import Expression
import copy
from coopr.pysp.phutils import update_all_rhos

from six import iteritems, advance_iterator

class DualPHModel(object):

    def _create_model(self):
        model = ConcreteModel()
        model.dual = Suffix(direction=Suffix.IMPORT,default=0.0)
        for stage in self._ph._scenario_tree._stages[:-1]: # all blended stages
            for tree_node in stage._tree_nodes:
                setattr(model,tree_node._name,Block())
                block = getattr(model,tree_node._name)
                block.var_to_id = {}
                block.id_to_var = []
                for cntr, (var,index) in enumerate((name,idx) for name, indices in iteritems(tree_node._variable_indices) for idx in indices):
                    block.var_to_id[var,index] = cntr
                    block.id_to_var.append((var,index))
                block.var_index = RangeSet(0,len(block.id_to_var)-1)
        self._model = model
        
    def __init__(self,ph):
        self._ph = ph
        self._model = None
        self._alphas = {}
        self._rhs = {}
        self._w_ips = {}
        self._wbars = {}
        self._iter = -1
        self._lhs = {}
        self._history = []
        self._create_model()

    def add_cut(self,first=False):
        self._iter += 1
        model = self._model
        model.del_component('cuts')
        model.cuts = RangeSet(0,self._iter)
        model.del_component('beta')
        model.beta = Var(model.cuts,within=NonNegativeReals)
        model.del_component('beta_sum_one')
        model.beta_sum_one = Constraint(expr=summation(model.beta)==1)
        lhs = 0.0
        for scenario in self._ph._scenario_tree._scenarios:
            instance = self._ph._instances[scenario._name]
            instance_objectives = instance.active_subcomponents(Objective)
            objective_name, objective = advance_iterator(iteritems(instance_objectives))
            lhs += -scenario._probability*value(instance_objectives[objective_name])
        self._lhs[self._iter] = lhs
        rhs = 0.0
        for stage in self._ph._scenario_tree._stages[:-1]:  # all blended stages
            for tree_node in stage._tree_nodes:
                for variable_name, variable_indices in iteritems(tree_node._variable_indices):
                    weight_parameter_name = "PHWEIGHT_"+variable_name
                    xbar = tree_node._xbars[variable_name]
                    for scenario in tree_node._scenarios:
                        instance = self._ph._instances[scenario._name]
                        weight_parameter = getattr(instance, weight_parameter_name)
                        for index in variable_indices:
                            rhs += scenario._probability*weight_parameter[index].value*xbar[index]
        self._rhs[self._iter] = rhs
        self._alphas[self._iter] = lhs+rhs
        model.del_component('obj')
        model.obj = Objective(expr=sum(self._alphas[i]*model.beta[i] for i in model.cuts))

        self._wbars[self._iter] = {}
        for stage in self._ph._scenario_tree._stages[:-1]: # all blended stages
            for tree_node in stage._tree_nodes:
                self._wbars[self._iter][tree_node._name] = copy.deepcopy(tree_node._wbars)
                block = getattr(model,tree_node._name)
                def _c_rule(block,i):
                    lhs = sum(model.beta[k]*self._wbars[k][tree_node._name][block.id_to_var[i][0]][block.id_to_var[i][1]] for k in model.beta.index_set())
                    if not isinstance(lhs,Expression):
                        return Constraint.Skip
                    return lhs == 0
                block.del_component('con')
                block.con = Constraint(block.var_index, rule=_c_rule)

    def solve(self):
        opt = SolverFactory("cplex")
        model = self._model
        model.dual.clearValue()
        model.preprocess()
        model.load(opt.solve(model))#,keepfiles=True,symbolic_solver_labels=True,tee=True))

        self._update_tree_node_xbars()

        self._w_ips[self._iter] = self._compute_weight_inner_product()

        print "**** Convergence Metric:", model.dual.getValue(model.beta_sum_one)-self._lhs[self._iter]
        self._history.append(model.dual.getValue(model.beta_sum_one)-self._lhs[self._iter])
        if ((model.dual.getValue(model.beta_sum_one)-self._lhs[self._iter]) < 1e-4) and (self._iter > 1):
            rho_old = self._ph._rho
            rho_new = self._ph._rho = rho_old*2
            if not (raw_input("Set rho to "+str(rho_new)+" ?") == 'n'):
                update_all_rhos(self._ph._instances, self._ph._scenario_tree, rho_value=self._ph._rho)
                #for i in xrange(len(self._alphas)):
                #    self._alphas[i] = self._alphas[i] + ((rho_new-rho_old) / 2.0) * self._w_ips[i]
                return False
            return True
        return False

    def _compute_weight_inner_product(self):
        w_inner_prod = 0.0
        for stage in self._ph._scenario_tree._stages[:-1]:  # all blended stages
            for tree_node in stage._tree_nodes:
                for variable_name, variable_indices in iteritems(tree_node._variable_indices):
                    weight_parameter_name = "PHWEIGHT_"+variable_name
                    for scenario in tree_node._scenarios:
                        instance = self._ph._instances[scenario._name]
                        weight_parameter = getattr(instance, weight_parameter_name)
                        for index in variable_indices:
                            w_inner_prod += weight_parameter[index].value*weight_parameter[index].value
        return w_inner_prod


    # This function updates the xbars dictionary on the ph scenario tree
    # to match the duals of the associated constraints in this model
    def _update_tree_node_xbars(self):
        # update the xbars
        model = self._model
        for stage in self._ph._scenario_tree._stages[:-1]: # all blended stages
            for tree_node in stage._tree_nodes:
                block = getattr(model,tree_node._name)
                for idx, con_data in iteritems(block.con):
                    var_name = block.id_to_var[idx][0]
                    var_index = block.id_to_var[idx][1]
                    tree_node._xbars[var_name][var_index] = model.dual.getValue(con_data)
